
template<typename T> class PointerTable
{
	template<typename Type> struct hashnode
	{
		Type data;
		hashnode<Type> *children;
		hashnode()
		{
			data.p = NULL;
			children = NULL;
		}
		~hashnode()
		{
			delete [] children;
		}
		void Clear()
		{
			delete [] children;
			data.p = NULL;
			children = NULL;
		}
	};
protected:
	static const int rootnodecount = 0x1000;
	hashnode<T> node[rootnodecount];
	intptr_t hash8(int32_t &ip)
	{
		int32_t w;
		w = ip / 127773;
		if((ip = (ip - w * 127773) * 16807 - w * 2836) <= 0) ip += 2147483647;
		return ip & 0x7;
	}
	intptr_t hash64(const uintptr_t uip)
	{
		return ((uip>>6) ^ (uip>>18)) & 0x3f;
	}
	intptr_t hash4096(const uintptr_t uip)
	{
		return ((uip<<6) ^ (uip>>12)) & 0xfff;
	}
	void Add8(hashnode<T> **n, T &x, int32_t ip)
	{
		intptr_t i;
		while(true){
			i = hash8(ip);
			if(*n == NULL){
				*n = new hashnode<T>[0x8];
			}
			if((*n)[i].data.p == NULL || (*n)[i].data.p == x.p){
				(*n)[i].data = x;
				break;
			}
			n = &(*n)[i].children;
		}
	}
	void Add64(hashnode<T> *&n, T &x)
	{
		intptr_t i = hash64((uintptr_t)x.p);
		if(n == NULL){
			n = new hashnode<T>[0x40];
		}
		if(n[i].data.p == NULL || n[i].data.p == x.p){
			n[i].data = x;
		}else{
			Add8(&n[i].children, x, (int32_t)(intptr_t)x.p);
		}
	}
	void Add(T &x)
	{
		if(x.p == NULL) return;
		intptr_t i = hash4096((uintptr_t)x.p);
		if(node[i].data.p == NULL || node[i].data.p == x.p){
			node[i].data = x;
		}else{
			Add64(node[i].children, x);
		}
	}
	void SetAtogama8(hashnode<T> &n)
	{
		static int si = 0;
		int i;
		if(n.children == NULL){
			n.data.p = NULL;
			return;
		}
		si++;
		for(i = 0; i < (0x8); i++){
			if(n.children[(si+i)&0x7].data.p != NULL){
				n.data = n.children[(si+i)&0x7].data;
				SetAtogama8(n.children[(si+i)&0x7]);
				return;
			}
		}
		n.Clear();
	}
	void SetAtogama64(hashnode<T> &n)
	{
		static int si = 0;
		int i;
		if(n.children == NULL){
			n.data.p = NULL;
			return;
		}
		si++;
		for(i = 0; i < (0x40); i++){
			if(n.children[(si+i)&0x3f].data.p != NULL){
				n.data = n.children[(si+i)&0x3f].data;
				SetAtogama8(n.children[(si+i)&0x3f]);
				return;
			}
		}
		n.Clear();
	}
	void Del8(hashnode<T> *n, void *const p, int32_t ip)
	{
		intptr_t i;
		while(n != NULL){
			i = hash8(ip);
			if(n[i].data.p == p){
				SetAtogama8(n[i]);
				break;
			}
			n = n[i].children;
		}
	}
	void Del64(hashnode<T> *const n, void *const p)
	{
		if(n == NULL) return;
		intptr_t i = hash64((uintptr_t)p);
		if(n[i].data.p == p){
			SetAtogama8(n[i]);
		}else{
			Del8(n[i].children, p, (int32_t)(intptr_t)p);
		}
	}
	void Del(void *const p)
	{
		if(p == NULL) return;
		intptr_t i = hash4096((uintptr_t)p);
		if(node[i].data.p == p){
			SetAtogama64(node[i]);
		}else{
			Del64(node[i].children, p);
		}
	}
	T* Get8(hashnode<T> *n, void *const p, int32_t ip)
	{
		intptr_t i;
		while(n != NULL){
			i = hash8(ip);
			if(n[i].data.p == p) return &n[i].data;
			n = n[i].children;
		}
		return NULL;
	}
	T* Get64(hashnode<T> *const n, void *const p)
	{
		if(n == NULL) return NULL;
		intptr_t i = hash64((uintptr_t)p);
		if(n[i].data.p == p) return &n[i].data;
		return Get8(n[i].children, p, (int32_t)(intptr_t)p);
	}
	T* Get(void *const p)
	{
		if(p == NULL) return NULL;
		intptr_t i = hash4096((uintptr_t)p);
		if(node[i].data.p == p) return &node[i].data;
		return Get64(node[i].children, p);
	}
};


struct FlagPTData
{
	void *p;
	bool flag;
	FlagPTData()
	{
		flag = false;
	}
};
class PointerFlagTable : public PointerTable<FlagPTData>
{
public:
	PWCArray<void*> memo;
	void Clear()
	{
		int i;
		for(i = 0; i < rootnodecount; i++) node[i].Clear();
		memo.Clear();
	}
	void set(void *p, bool f = true)
	{
		FlagPTData fn;
		fn.p = p;
		fn.flag = f;
		Add(fn);
		memo += p;
	}
	bool get(void *p)
	{
		FlagPTData *n = Get(p);
		if(n == NULL) return false;
		return n->flag;
	}
};

struct JunkanPTData
{
	void *p;
	intptr_t rpid;
	bool kanousei;
};
class JunkanGoroshi : public PointerTable<JunkanPTData>
{
	enum RestartPoint
	{
		P_FOOBAR, P_OOF, P_ZAB
	};
	struct TansakuLoopInfo
	{
		HeapObj *hp, *rthp;
		intptr_t rpid, i, j, rtj, rtn;
		RestartPoint rp;
	};
	PWCArrayQuad< PWCArray<intptr_t> > refposlist;
	PWCReverseStack<HeapObj*> queue;
	LockSingler mtx;
	bool JunkanCheck(PWCArray<TansakuLoopInfo> &LIStack, TansakuLoopInfo &inf, const RestartPoint rp, HeapObj *const hobj, const intptr_t id, PointerFlagTable &pft, uint32_t &ncnt, uint32_t &rcnt, uint32_t &jcnt, HeapObj *const start)
	{
		ncnt++;
		if(pft.get(hobj)){
			if(hobj == start) jcnt++;
			return false;
		}
		pft.set(hobj);
		rcnt += hobj->head.refcount;
		inf.rp = rp;
		LIStack += inf;
		inf.hp = hobj;
		inf.rpid = id;
		return true;
	}
public:
	SourceTree<JunkanGoroshi> *psot;
	const PWCArray<uint8_t> *pbcode;
	PWCArray<int8_t> *pgres;
	void JunkanTansaku(HeapObj *const hobj, const intptr_t id, PointerFlagTable &pft, uint32_t &ncnt, uint32_t &rcnt, uint32_t &jcnt, HeapObj *const start)
	{
		PWCArray<TansakuLoopInfo> LIStack;
		TansakuLoopInfo inf;
		HeapObj *ph;
		DynamicRef *dypd;
		intptr_t rtid;
		int32_t typ;
		pft.Clear();
		ncnt = rcnt = jcnt = 0;
		JunkanCheck(LIStack, inf, P_FOOBAR, hobj, id, pft, ncnt, rcnt, jcnt, start);
		LIStack.Clear();
RABOOF:
		for(inf.j = 0; inf.j < inf.hp->head.datasize; inf.j += AtPL(inf.rpid).GetConst()[0]){
			for(inf.i = 2; inf.i < AtPL(inf.rpid).GetLength(); inf.i += 3){
				if(AtPL(inf.rpid).GetConst()[inf.i+1] < 0 && AtPL(inf.rpid).GetConst()[inf.i+2] == 0){
					dypd = (DynamicRef*)(inf.hp->head.data+inf.j+AtPL(inf.rpid).GetConst()[inf.i]);
					ph = dypd->obj.pointer;
					typ = dypd->typ;
					if(ph == dypd->obj.pointer && typ == dypd->typ &&
						(intptr_t)ph > 0 && 0 <= typ && typ <= ENDCLASS_TYPEID && psot->stat->funclist.Get(typ)->IsJunkanAble()){
						if(JunkanCheck(LIStack, inf, P_FOOBAR, ph, psot->stat->funclist.Get(typ)->sonota.junkanid, pft, ncnt, rcnt, jcnt, start)) goto RABOOF;
					}
				}else{
					inf.rthp = ((Reference*)(inf.hp->head.data+inf.j+AtPL(inf.rpid).GetConst()[inf.i]))->pointer;
					if((intptr_t)inf.rthp > 0){
						inf.rtn = AtPL(inf.rpid).GetConst()[inf.i+2];
FOO:
						rtid = psot->stat->funclist.Get((int32_t)AtPL(inf.rpid).GetConst()[inf.i+1])->sonota.junkanid;
						if(inf.rtn <= 0){
							if(rtid < 0){
								for(inf.rtj = 0; inf.rtj < inf.rthp->head.datasize; inf.rtj += sizeof(DynamicRef)){
									dypd = (DynamicRef *)(inf.rthp->head.data+inf.rtj);
									ph = dypd->obj.pointer;
									typ = dypd->typ;
									if(ph == dypd->obj.pointer && typ == dypd->typ &&
										(intptr_t)ph > 0 && 0 <= typ && dypd->typ <= ENDCLASS_TYPEID && psot->stat->funclist.Get(typ)->IsJunkanAble()){
										if(JunkanCheck(LIStack, inf, P_ZAB, ph, psot->stat->funclist.Get(typ)->sonota.junkanid, pft, ncnt, rcnt, jcnt, start)) goto RABOOF;
									}
ZAB:;
								}
							}else{
								if(JunkanCheck(LIStack, inf, P_FOOBAR, inf.rthp, rtid, pft, ncnt, rcnt, jcnt, start)) goto RABOOF;
							}
						}else{
							for(inf.rtj = 0; inf.rtj < inf.rthp->head.datasize; inf.rtj += sizeof(Reference)){
								ph = ((Reference *)(inf.rthp->head.data+inf.rtj))->pointer;
								if((intptr_t)ph > 0){
									inf.rp = P_OOF;
									LIStack += inf;
									inf.rthp = ph;
									inf.rtn--;
									goto FOO;
								}
OOF:;
							}
						}
					}
				}
FOOBAR:;
			}
		}
		if(LIStack.GetLength() > 0){
			inf = LIStack.Pop();
			switch(inf.rp){
			case P_FOOBAR:
				goto FOOBAR;
			case P_OOF:
				goto OOF;
			case P_ZAB:
				goto ZAB;
			}
		}
	}
	void Clear()
	{
		int i;
		refposlist.Clear();
		queue.Clear();
		for(i = 0; i < rootnodecount; i++) node[i].Clear();
	}
	void SetKankyou(SourceTree<JunkanGoroshi> *const st, const PWCArray<uint8_t> *const pb, PWCArray<int8_t> *const pg)
	{
		psot = st;
		pbcode = pb;
		pgres = pg;
	}
	intptr_t AddPL(const PWCArray<intptr_t> &rp)
	{
		refposlist += rp;
		return refposlist.GetLength()-1;
	}
	void SetPLDestructor(const intptr_t i, const intptr_t des)
	{
		refposlist.GetData()[i].GetData()[1] = des;
	}
	const PWCArray<intptr_t>& AtPL(const intptr_t i)
	{
		return refposlist.GetConst()[i];
	}
	bool PopNode(HeapObj *&p, intptr_t &rpid)
	{
		AutoLocker al(&mtx);
		JunkanPTData *n;
		for(;;){
			if(queue.IsEmpty()) return false;
			n = Get(queue.Pop());
			if(n == NULL) continue;
			if(n->kanousei) break;
		}
		p = (HeapObj*)n->p;
		rpid = n->rpid;
		n->kanousei = false;
		InterlockedIncrement(&g_KaihouStopCount);
		return true;
	}
	uint32_t KnouGet(HeapObj **const p, intptr_t l)
	{
		AutoLocker al(&mtx);
		intptr_t i;
		uint32_t rc = 0;
		for(i = 0; i < l; i++){
			rc += p[i]->head.refcount;
			JunkanPTData *n = Get(p[i]);
			if(n == NULL || n->kanousei) return 0;
		}
		return rc;
	}
	void RefNew(const intptr_t rpid, HeapObj *const p)
	{
		if(p == NULL) return;
		AutoLocker al(&mtx);
		intptr_t i = hash4096((uintptr_t)p);
		if(node[i].data.p == p) return;
		if(node[i].data.p == NULL){
			node[i].data.rpid = rpid;
			node[i].data.p = p;
		}else{
			JunkanPTData jn;
			jn.p = p;
			jn.rpid = rpid;
			jn.kanousei = false;
			Add64(node[i].children, jn);
		}
	}
	void RefDel(HeapObj *const p)
	{
		AutoLocker al(&mtx);
		Del(p);
	}
	void RefDwn(HeapObj *const p)
	{
		AutoLocker al(&mtx);
		JunkanPTData *n = Get(p);
		if(n == NULL) return;
		if(!n->kanousei){
			n->kanousei = true;
			queue.Push(p);
		}
	}
};

