struct HeapObjHead
{
	int8_t *data;
	intptr_t datasize;
	HANDLE mutex;
	uint32_t refcount;
};
struct HeapObj
{
	HeapObjHead head;
	union{
		int8_t data[1];
		intptr_t limitsize;
	} body;
	void createmutex()
	{
		if(InterlockedCompareExchange((LPLONG)&head.mutex, ~NULL, NULL) != NULL){
			while(head.mutex == (HANDLE)~NULL) Sleep(0);
			return;
		}
		head.mutex = CreateMutex(NULL, FALSE, NULL);
	}
	void lock()
	{
		createmutex();
		WaitForSingleObject(head.mutex, INFINITE);
	}
	void unlock()
	{
		ReleaseMutex(head.mutex);
	}
	int8_t* at(intptr_t i, const intptr_t tsize) const
	{
		intptr_t le, tmp;
		le = (uintptr_t)head.datasize / (uintptr_t)tsize;
		tmp = (i>>31) | (((le-1)-i)>>31);
		i &= (~tmp & i) | (tmp & (le-1));
		return head.data + i*tsize;
	}
	template<intptr_t tsize> int8_t* at(intptr_t i) const
	{
		intptr_t le, tmp;
		le = (uintptr_t)head.datasize / (uintptr_t)tsize;
		tmp = (i>>31) | (((le-1)-i)>>31);
		i &= (~tmp & i) | (tmp & (le-1));
		return head.data + i*tsize;
	}
	void delet()
	{
		if(head.data != body.data) sszrefdeletefunc(head.data);
		if(head.mutex != NULL) CloseHandle(head.mutex);
		sszrefdeletefunc(this);
	}
};
struct Reference
{
	HeapObj *pointer;
	intptr_t position;
	intptr_t length;
	void init()
	{
		pointer = NULL;
		position = length = 0;
	}
	bool null() const
	{
		return pointer == NULL;
	}
	HeapObj* obj() const
	{
		return (HeapObj*)((intptr_t)pointer & INTPTR_MAX);
	}
	intptr_t pos() const
	{
		return position;
	}
	intptr_t len() const
	{
		if(length < 0){
			return obj()->head.datasize;
		}
		return length;
	}
	intptr_t size() const
	{
		if(null()) return 0;
		return obj()->head.datasize;
	}
	uint32_t refcnt() const
	{
		if(null()) return 0;
		return obj()->head.refcount;
	}
	void countup()
	{
		if(null()) return;
		InterlockedIncrement((LONG*)&obj()->head.refcount);
	}
	HeapObj* release()
	{
		HeapObj *p = NULL;
		if((intptr_t)pointer > 0 && InterlockedDecrement((LPLONG)&pointer->head.refcount) == 0) p = pointer;
		init();
		return p;
	}
	void releaseanddelete()
	{
		HeapObj *p;
		if(null()) return;
		p = release();
		if(p != NULL) p->delet();
	}
	HeapObj* copy(const Reference &r)
	{
		HeapObj *p;
		if(r.null()) return release();
		p = pointer;
		*this = r;
		if(obj() == (HeapObj*)((intptr_t)p & INTPTR_MAX)){
			if((intptr_t)p > 0){
				if((intptr_t)pointer > 0 || InterlockedDecrement((LPLONG)&p->head.refcount) != 0) p = NULL;
			}else{
				if((intptr_t)pointer > 0) InterlockedIncrement((LONG*)&p->head.refcount);
				p = NULL;
			}
			return p;
		}else if((intptr_t)p > 0){
			if(InterlockedDecrement((LPLONG)&p->head.refcount) != 0) p = NULL;
		}else{
			p = NULL;
		}
		if((intptr_t)pointer > 0) InterlockedIncrement((LONG*)&pointer->head.refcount);
		return p;
	}
	HeapObj* sref(const Reference &r)
	{
		HeapObj *p;
		if(r.null()) return release();
		p = pointer;
		*this = r;
		*(intptr_t*)&pointer &= INTPTR_MAX;
		if(pointer == (HeapObj*)((intptr_t)p & INTPTR_MAX)){
			if((intptr_t)p < 0) InterlockedIncrement((LONG*)&p->head.refcount);
			return NULL;
		}else if((intptr_t)p > 0){
			if(InterlockedDecrement((LPLONG)&p->head.refcount) != 0) p = NULL;
		}else{
			p = NULL;
		}
		InterlockedIncrement((LONG*)&obj()->head.refcount);
		return p;
	}
	HeapObj* wref(const Reference &r)
	{
		HeapObj *p;
		if(r.null()) return release();
		p = pointer;
		*this = r;
		*(intptr_t*)&pointer |= INTPTR_MIN;
		if(obj() == (HeapObj*)((intptr_t)p & INTPTR_MAX)){
			if((intptr_t)p > 0){
				if((intptr_t)pointer > 0 || InterlockedDecrement((LPLONG)&p->head.refcount) != 0) p = NULL;
			}else{
				p = NULL;
			}
			return p;
		}else if((intptr_t)p > 0){
			if(InterlockedDecrement((LPLONG)&p->head.refcount) != 0) p = NULL;
		}else{
			p = NULL;
		}
		return p;
	}
	void lock()
	{
		if(null()) return;
		obj()->lock();
	}
	void unlock()
	{
		if(null()) return;
		obj()->unlock();
	}
//////////////////////////////////////////////////////////////////////////////
	int8_t* atpos() const
	{
		return obj()->head.data + position;
	}
	int8_t* refat(intptr_t i, const intptr_t tsize) const
	{
		intptr_t le, tmp;
		if(length < 0) return listat(i, tsize);
		le = (uintptr_t)length / (uintptr_t)tsize;
		tmp = (i>>31) | (((le-1)-i)>>31);
		i &= (~tmp & i) | (tmp & (le-1));
		return obj()->head.data+position+i*tsize;
	}
	template<intptr_t tsize> int8_t* refat(intptr_t i) const
	{
		intptr_t le, tmp;
		if(length < 0) return listat<tsize>(i);
		le = (uintptr_t)length / (uintptr_t)tsize;
		tmp = (i>>31) | (((le-1)-i)>>31);
		i &= (~tmp & i) | (tmp & (le-1));
		return obj()->head.data+position+i*tsize;
	}
	int8_t* listat(const intptr_t i, const intptr_t tsize) const
	{
		return obj()->at(i, tsize);
	}
	template<intptr_t tsize> int8_t* listat(const intptr_t i) const
	{
		return obj()->at<tsize>(i);
	}
	int8_t* listnewat(const intptr_t i, const intptr_t tsize, void (__stdcall *fp)(void*, int8_t*, intptr_t) = NULL, void *p = NULL)
	{
		intptr_t idx;
		if(i < 0){
			idx = addsize(-i, tsize, fp, p);
			return obj()->head.data + idx;
		}
		if(null()){
			listnew(i+1, tsize, fp, p);
		}else if((i+1)*tsize > obj()->head.datasize){
			addsize(((i+1)*tsize - obj()->head.datasize)/tsize, tsize, fp, p);
		}
		return obj()->head.data + i*tsize;
	}
	void refnew(const intptr_t s, const intptr_t tsize, void (__stdcall *fp)(void*, int8_t*, intptr_t) = NULL, void *p = NULL)//GCXbhɒ
	{
		if(s*tsize <= 0) return;
		HeapObj *tmp = (HeapObj*)sszrefnewfunc(sizeof(HeapObjHead)+s*tsize);
		tmp->head.data = tmp->body.data;
		tmp->head.datasize = s*tsize;
		tmp->head.mutex = NULL;
		tmp->head.refcount = 1;
		if(fp != NULL) fp(p, tmp->head.data, s*tsize);
		pointer = tmp;
		position = 0;
		length = s*tsize;
	}
	void listnew(const intptr_t s, const intptr_t tsize, void (__stdcall *fp)(void*, int8_t*, intptr_t) = NULL, void *p = NULL)//GCXbhɒ
	{
		if(s*tsize <= 0) return;
		HeapObj *tmp = (HeapObj*)sszrefnewfunc(sizeof(HeapObj));
		tmp->head.data = (int8_t*)sszrefnewfunc(s*tsize);
		tmp->head.datasize = tmp->body.limitsize = s*tsize;
		tmp->head.mutex = NULL;
		tmp->head.refcount = 1;
		if(fp != NULL) fp(p, tmp->head.data, s*tsize);
		pointer = tmp;
		position = 0;
		length = -1;
	}
	intptr_t addsize(const intptr_t s, const intptr_t tsize, void (__stdcall *fp)(void*, int8_t*, intptr_t) = NULL, void *p = NULL)//GCXbhɒ
	{
		intptr_t tmps, ds;
		int8_t *tmpd;
		if(null()){
			listnew(s, tsize, fp, p);
			return 0;
		}
		tmps = obj()->head.datasize;
		if(s*tsize <= 0) return tmps;
		ds = obj()->head.datasize + s*tsize;
		if(ds > obj()->body.limitsize){
			obj()->body.limitsize = ds + (ds >> 2);
			tmpd = (int8_t*)sszrefnewfunc(obj()->body.limitsize);
			memcpy(tmpd, obj()->head.data, tmps);
			sszrefdeletefunc(obj()->head.data);
			if(fp != NULL) fp(p, tmpd+tmps, s*tsize);
			obj()->head.data = tmpd;
		}else{
			if(fp != NULL) fp(p, obj()->head.data+tmps, s*tsize);
		}
		obj()->head.datasize = ds;
		return tmps;
	}
	void slice(intptr_t kara, intptr_t made, const intptr_t tsize)
	{
		intptr_t le, tmp;
		le = (uintptr_t)len() / (uintptr_t)tsize;
		tmp = (kara>>31) | ((le-kara)>>31);
		kara = (~tmp & kara) | (tmp & le);
		tmp = (made>>31) | ((le-made)>>31);
		made = (~tmp & made) | (tmp & le);
		le = made - kara;
		tmp = ~((le>>31) | ((le-1)>>31));
		*(intptr_t*)&pointer &= tmp;
		position = tmp & (position + kara*tsize);
		length = (tmp & le)*tsize;
	}
	template<intptr_t tsize> void slice(intptr_t kara, intptr_t made)
	{
		intptr_t le, tmp;
		le = (uintptr_t)len() / (uintptr_t)tsize;
		tmp = (kara>>31) | ((le-kara)>>31);
		kara = (~tmp & kara) | (tmp & le);
		tmp = (made>>31) | ((le-made)>>31);
		made = (~tmp & made) | (tmp & le);
		le = made - kara;
		tmp = ~((le>>31) | ((le-1)>>31));
		*(intptr_t*)&pointer &= tmp;
		position = tmp & (position + kara*tsize);
		length = (tmp & le)*tsize;
	}
};
struct DynamicRef
{
	Reference obj;
	int32_t typ;
	void init()
	{
		typ = VOID_TYPEID;
		obj.init();
	}
};
void __stdcall refzeroclearcb(void* p, int8_t* pd, intptr_t l)
{
	ZeroMemory(pd, l);
}
void __stdcall refdynnullclearcb(void* p, int8_t* pd, intptr_t l)
{
	intptr_t i;
	for(i = 0; i < l; i += sizeof(DynamicRef)){
		((DynamicRef*)(pd+i))->init();
	}
}

struct Thread
{
	HANDLE handle;
	void init()
	{
		handle = NULL;
	}
	bool null() const
	{
		return handle == NULL;
	}
	void wait()
	{
		if(null()) return;
		WaitForSingleObject(handle, INFINITE);
		CloseHandle(handle);
		handle = NULL;
	}
	void create(unsigned (__stdcall *const lsr)(void*), void *const param)
	{
		wait();
		handle = (HANDLE)_beginthreadex(NULL, 0, lsr, param, 0, NULL);
	}
};
